DROP TABLE AirAirlineSource;
DROP TABLE AirAirportModesSource;
DROP TABLE AirAirportNamesSource;
DROP TABLE AirAirportSource;
DROP TABLE AirGateSource;
DROP TABLE AirFlightSource;
DROP TABLE BusCompanySource;
DROP TABLE BusLineSource;
DROP TABLE BusStopSource;
DROP TABLE BusBerthSource;
DROP TABLE BusConnectionSource;
DROP TABLE RailCompanySource;
DROP TABLE RailLineSource;
DROP TABLE RailStationSource;
DROP TABLE RailPlatformSource;
DROP TABLE RailConnectionSource;
DROP TABLE SeaCompanySource;
DROP TABLE SeaLineSource;
DROP TABLE SeaStopSource;
DROP TABLE SeaDockSource;
DROP TABLE SeaConnectionSource;
DROP TABLE ProximitySource;
DROP TABLE NodeLocationSource;
DROP TABLE NodeSource;
DROP TABLE Source;
UPDATE Metadata SET has_sources = false WHERE ROWID = 1;
VACUUM;